// state.h

#define _CRT_SECURE_NO_WARNINGS

#ifndef STATE_H
#define STATE_H

#ifndef NDEBUG
    #define NDEBUG
#endif

/*
 *	DATA.C
 *	Tom Kerrigan's Simple Chess Program (TSCP)
 *
 *	Copyright 1997 Tom Kerrigan
 */

#include "defs.h"
#include <string>
#include <iostream>
#include <vector>
#include <chrono>
#include <assert.h>

/* the board representation */
//int color[64];  /* WHITE, BLACK, or EMPTY */
//int piece[64];  /* PAWN, KNIGHT, BISHOP, ROOK, QUEEN, KING, or EMPTY */
//int side;  /* the side to move */
//int xside;  /* the side not to move */
//int castle;  /* a bitfield with the castle permissions. if 1 is set,
//                white can still castle kingside. 2 is white queenside.
//				4 is black kingside. 8 is black queenside. */
//int ep;  /* the en passant square. if white moves e2e4, the en passant
//            square is set to e3, because that's where a pawn would move
//			in an en passant capture */
//int fifty;  /* the number of moves since a capture or pawn move, used
//               to handle the fifty-move-draw rule */
//int hash;  /* a (more or less) unique number that corresponds to the
//              position */
//int ply;  /* the number of half-moves (ply) since the
//             root of the search tree */
//int hply;  /* h for history; the number of ply since the beginning
//              of the game */

/* gen_dat is some memory for move lists that are created by the move
   generators. The move list for ply n starts at first_move[n] and ends
   at first_move[n + 1]. */
//gen_t gen_dat[GEN_STACK];
//int first_move[MAX_PLY];

///* the history heuristic array (used for move ordering) */
//int history[64][64];

///* we need an array of hist_t's so we can take back the
//   moves we make */
//hist_t hist_dat[HIST_STACK];

///* the engine will search for max_time milliseconds or until it finishes
//   searching max_depth ply. */
//int max_time;
//int max_depth;

///* the time when the engine starts searching, and when it should stop */
//int start_time;
//int stop_time;

//int nodes;  /* the number of nodes we've searched */

///* a "triangular" PV array; for a good explanation of why a triangular
//   array is needed, see "How Computers Play Chess" by Levy and Newborn. */
//move pv[MAX_PLY][MAX_PLY];
//int pv_length[MAX_PLY];
//BOOL follow_pv;

///* random numbers used to compute hash; see set_hash() in board.c */
//int hash_piece[2][6][64];  /* indexed by piece [color][type][square] */
//int hash_side;
//int hash_ep[64];

//enum {
//    A8 =  0, B8, C8, D8, E8, F8, G8, H8,
//    A7 =  8, B7, C7, D7, E7, F7, G7, H7,
//    A6 = 16, B6, C6, D6, E6, F6, G6, H6,
//    A5 = 24, B5, C5, D5, E5, F5, G5, H5,
//    A4 = 32, B4, C4, D4, E4, F4, G4, H4,
//    A3 = 40, B3, C3, D3, E3, F3, G3, H3,
//    A2 = 48, B2, C2, D2, E2, F2, G2, H2,
//    A1 = 56, B1, C1, D1, E1, F1, G1, H1
//};

enum {NORESULT=0,DRAW=1,MATE=2, WINNING=3};
enum {SOUTH=0, NORTH=1, BOTH=3, NONE=4};
enum { HFNONE, HFALPHA, HFBETA, HFEXACT };

enum {sP = 0, sN, sB, sR, sQ, sK, nP = 7, nN, nB, nR, nQ, nK};
enum { FILE_A, FILE_B, FILE_C, FILE_D, FILE_E, FILE_F, FILE_G, FILE_H, FILE_NONE };
enum { RANK_8, RANK_7, RANK_6, RANK_5, RANK_4, RANK_3, RANK_2, RANK_1, RANK_NONE };
enum { WKCA = 1, WQCA = 2, BKCA = 4, BQCA = 8 };

void init_tables();

const int SIM_DEPTH = 2;
const double UCB_CONST = 1.4;
const int DEFAULT_TIME = 5000;

class State
{
    friend class Node;
    friend umove searchMC(State *rootstate, int niterations, std::chrono::milliseconds timelimit);

public:
    State();
    // copy constructor
    State(const State& old);
    // assignment operator
    State& operator=(const State& source);
    // destructor
    ~State() {}

public:
	/* board.c */
	void init_board();
    bool readFen(std::string fen);
	BOOL in_check(int s);
	BOOL attack(int sq, int s);
	void gen();
	void gen_caps();
	void gen_push(int from, int to, int bits);
	void gen_promote(int from, int to, int bits);
    umove getGenDat(int i) const { return gen_dat[i].m;}
    int genLegal(std::vector<umove> &list);
    BOOL makemove(move_bytes m);
	void takeback();
    void set_hash();

	/* book.c */
	int book_move();
	BOOL book_match(char *s1, char *s2);

	/* search.c */
    double simulation();
	void think(int output);
	int search(int alpha, int beta, int depth);
	int quiesce(int alpha, int beta);
	int reps();
	void sort_pv();
	void sort(int from);
	void checkup();

	/* eval.c */
	int eval();

    /* utility.c */
    move_bytes sudoMove(int f, int t);
    bool sudoMove(int f, int t, move_bytes& mb);
    int parse_move(char *s);
	char *move_str(move_bytes m);
	void print_board();

    // Misc
    int position(int sq) const { return piece[sq] + (color[sq] == 1? 7:0);}
    int getColor(int sq) const { return color[sq];}
    int getSide() const { return side;}
    int getEp() const { return ep;}
    int getHPly() const { return hply;}
    void resetPly() { ply = 0;}
    bool isGameOver();
    umove getPV() const { return pv[0][0];} // edited
    void printStatus();
    int moveCount() const { return first_move[ply+1] - first_move[ply];}
    void setSearchTime(int ms) {if (ms >= 1000 && ms <= 5000) max_time = ms; else ms = 1000;}
    void print_result();
    void bench();
    void xboard();
private:
	int color[64];
	int piece[64];
	int kings[2];
	int side;
	int xside;
	int castle;
	int ep;
	int fifty;
	int hash;
	int ply;
	int hply;
	gen_t gen_dat[GEN_STACK];
	int first_move[MAX_PLY];
	int history[64][64];
	hist_t hist_dat[HIST_STACK];
	int max_time;
	int max_depth;
	int start_time;
	int stop_time;
	int nodes;
    umove pv[MAX_PLY][MAX_PLY]; // edited
	int pv_length[MAX_PLY];
	BOOL follow_pv;
};

// utility.cpp
extern 	int get_ms();
extern void init_hash();
extern int hash_rand();
extern void open_book();
extern void close_book();

// const.cpp
extern const int mailbox[120];
extern const int mailbox64[64];
extern const BOOL slide[6];
extern const int offsets[6];
extern const int offset[6][8];
extern const int castle_mask[64];
extern const char piece_char[6];
extern const int init_color[64];
extern const int init_piece[64];
extern const int init_empty[64];
extern const int NOSQ;
extern int hash_piece[2][6][64];
extern int hash_side;
extern int hash_ep[64];

extern int mg_table[2][6][64];
extern int eg_table[2][6][64];

// mcts.cpp
extern umove searchMC(State *rootstate, int niterations, std::chrono::milliseconds timelimit);

// utility.cpp
extern double sigmoid(int score);
extern std::string move2str(move_bytes mb);

#endif // STATE_H
